/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Spicebird code.
 *
 * The Initial Developer of the Original Code is
 * Synovel Software Technologies
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Prasad Sunkari <prasad@synovel.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef _csTPTextChannel_H_
#define _csTPTextChannel_H_

#include "csITPTextChannel.h"
#include "csTPChannel.h"
#include "nsStringGlue.h"

/* Header file */
class csTPTextChannel : public csITPTextChannel, 
                        public csTPChannel
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_CSITPTEXTCHANNEL

  csTPTextChannel();

  void LostMessage();
  void MessageReceived(PRUint32 aId, PRUint32 aTimeStamp, PRUint32 aSender,
                       PRUint32 aType, PRUint32 aFlags, const gchar *aMessage);
  void SendError(PRUint32 aError, PRUint32 aTimeStamp,
                 PRUint32 aType, const gchar *aMessage);
  void MessageSent(PRUint32 aTimeStamp,
                   PRUint32 aType, const gchar *aMessage);

private:
  ~csTPTextChannel();

protected:
  nsCOMPtr<nsIMutableArray> m_LostMessageObservers;
  nsCOMPtr<nsIMutableArray> m_ReceivedObservers;
  nsCOMPtr<nsIMutableArray> m_SendErrorObservers;
  nsCOMPtr<nsIMutableArray> m_SentObservers;
};

class csTPTextChannelPendingMessage : public csITPTextChannelPendingMessage
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_CSITPTEXTCHANNELPENDINGMESSAGE

  csTPTextChannelPendingMessage(PRUint32 aId, PRUint32 aTimeStamp,
                                PRUint32 aSenderHandle, PRUint32 aMessageType,
                                PRUint32 aMessageFlags, nsString aMessageString);

  static NS_METHOD Create(PRUint32 aId, PRUint32 aTimeStamp, PRUint32 aSenderHandle,
                          PRUint32 aMessageType, PRUint32 aMessageFlags,
                          nsString aMessageString, void **aResult);
private:
  ~csTPTextChannelPendingMessage();

protected:
  PRUint32 m_Id;
  PRUint32 m_TimeStamp;
  PRUint32 m_SenderHandle;
  PRUint32 m_MessageType;
  PRUint32 m_MessageFlags;
  nsString m_MessageString;
};

#endif
